/*
Caterkiller
	by Lach
*/

local SKIN = "caterkiller"

local caterkillercollide = CV_RegisterVar( // you can turn off the body part bumping if you want to play caterkiller seriously for whatever reason
{
	name = "caterkillercollide",
	PossibleValue = CV_OnOff,
	defaultvalue = "On",
	flags = CV_NETVAR
})

freeslot("SPR_CKIP", "MT_CK_PART", "S_CK_PART", "S_CK_PART_PAIN")

states[S_CK_PART] = {
	sprite = SPR_CKIP
}

states[S_CK_PART_PAIN] = {
	sprite = SPR_CKIP,
	tics = 24,
	action = A_Scream
}

mobjinfo[MT_CK_PART] = {
	spawnstate = S_CK_PART,
	radius = mobjinfo[MT_PLAYER].radius/2,
	height = mobjinfo[MT_PLAYER].height/2,
	flags = MF_SPECIAL | MF_BOUNCE | MF_DONTENCOREMAP,
	activesound = sfx_s3k49,
	spawnhealth = 1,
	deathstate = S_CK_PART_PAIN,
	deathsound = sfx_s3k5d
}

//body part mechanics

local function killCaterkillerPart(mo)
	P_KillMobj(mo) // (this "kills" body parts and sets them to their deathstate)
	mo.movedir = (2*P_RandomKey(2) - 1)*ANG20
	mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_GRENADEBOUNCE) | MF_NOCLIPTHING // Make them collide with geometry, it's fun as fuck >u>
	P_InstaThrust(mo, mo.tracer.angle + P_RandomRange(-20, 20)*ANG1, FixedHypot(mo.tracer.momx, mo.tracer.momy) + 8*mo.tracer.scale) // fling them away
	mo.momz = P_MobjFlip(mo.tracer)*10*mo.scale // bit of vertical momentum too
end

local function pullPart(mo, target, length, hdist, factor) // This function is basically just tether physics but with everything that makes it tether physics stripped out
	local angle = R_PointToAngle2(mo.x, mo.y, target.x, target.y)
	local zangle = R_PointToAngle2(0, mo.z, hdist, target.z)
	
	local goal = {
		["x"] = target.x - FixedMul(length, FixedMul(cos(angle), cos(zangle))),
		["y"] = target.y - FixedMul(length, FixedMul(sin(angle), cos(zangle))),
		["z"] = target.z - FixedMul(length, sin(zangle)) + (mo.eflags & MFE_VERTICALFLIP and mo.height or 0)
	}
	
	local mass = 0
	
	local vars = {"x", "y", "z"}
	
	for i = 1, #vars
		local v = vars[i]
		
		mo["mom"..v] = $ + (goal[v] - mo[v] - $)/factor
		
		/*local limit = length + abs(target["mom"..v])
		mo["mom"..v] = max(min(limit, $ + FixedMul(FRACUNIT - mass, goal[v] - mo[v])), -limit)
		if target.type == mo.type
			target["mom"..v] = max(min(limit, $ + FixedMul(mass, mo[v] - goal[v])), -limit)
		end*/
	end
end

addHook("MobjThinker", function(mo) // body part thinker
	if mo and mo.valid
		if not (mo.target and mo.target.valid
		and mo.tracer and mo.tracer.valid and mo.tracer.skin == SKIN
		and mo.tracer.player)
			P_RemoveMobj(mo) // Remove body part if not attached to a caterkiller player
			return
		end
		
		local player = mo.tracer.player
		
		if not mo.ck_shadow
			P_SpawnShadowMobj(mo)
			mo.ck_shadow = true
		end
		
		if mo.pmomz and not P_IsObjectOnGround(mo)
			mo.pmomz = 0
		end
		
		if mo.tracer.state >= S_KART_SPIN and mo.tracer.state <= S_KART_PAIN and mo.health // if the player spins out, drop their balls :sunglasses:
			killCaterkillerPart(mo)
		end
		
		if not mo.health // assume body part is in deathstate
			mo.angle = $ + mo.movedir // spin the body part!
			if mo.tics % 2 // flash like the rocket sneakers
				mo.flags2 = $ & ~MF2_DONTDRAW
			else
				mo.flags2 = $ | MF2_DONTDRAW
			end
			
			return // leave the thinker, we don't need to run more code for dead body parts
		end
		
		if player.rmomx or player.rmomy // use slippery friction
			mo.friction = 63*FRACUNIT/64
		end
		
		if player.powers[pw_flashing] // no bumping when intangible!
			mo.flags = $ | MF_NOCLIPTHING
		else
			mo.flags = $ & ~MF_NOCLIPTHING
		end
		
		// set frame
		// (this makes me want switch case in Lua, there's probably a better way to do this anyway but)
		
		if mo.tracer.state >= S_KART_STND1 and mo.tracer.state <= S_KART_STND2_R // idle
			mo.frame = mo.tracer.state % 2 == 0 and B or A
		elseif mo.tracer.state >= S_KART_WALK1 and mo.tracer.state <= S_KART_WALK2_R // slow
			mo.frame = mo.tracer.state % 2 and C or D
		elseif mo.tracer.state >= S_KART_DRIFT1_L and mo.tracer.state <= S_KART_DRIFT2_R // drift
			mo.frame = mo.tracer.state % 2 and D or C
		elseif player.panim == PA_RUN // fast
			mo.frame = mo.tracer.state % 2 == 0 and C or A 
		elseif mo.tracer.state >= S_KART_SPIN and mo.tracer.state <= S_KART_PAIN // hurt
			mo.frame = A
		elseif mo.tracer.state == S_KART_SQUISH //squished
			mo.frame = E
		end
		
		mo.frame = $ | (mo.tracer.frame & ~FF_FRAMEMASK) // get FF_FULLBRIGHT and FF_TRANS flags applied
		mo.scale = mo.tracer.scale // set to player's scale
		mo.color = mo.tracer.color // set to player's color
		mo.colorized = mo.tracer.colorized // set to player's colorization
		mo.flags2 = mo.tracer.flags2 // get MF2_DONTDRAW or MF2_OBJECTFLIP etc.
		mo.eflags = mo.tracer.eflags // get MFE_VERTICALFLIP or MFE_ splitscreen flags etc.
		
		if player.kartstuff[k_invincibilitytimer] // spawn body part ghosts under certain circumstances
			P_SpawnGhostMobj(mo).fuse = 4
		elseif (player.kartstuff[k_driftboost]
		or player.kartstuff[k_sneakertimer]
		or player.kartstuff[k_startboost])
			P_SpawnGhostMobj(mo).fuse = 2
		end
		
		//Here's the fun bit!
		
		local length = 4*mo.radius // max distance between two parts before they are pulled
		
		local target = mo.target
		
		while target and target.valid // iterate through the other body parts and push away for each one that is too close
			local hdist = FixedHypot(mo.x - target.x, mo.y - target.y)
			local dist = FixedHypot(hdist, abs(mo.z - target.z))
			
			if dist < length/2
				pullPart(mo, target, length/2, hdist, player.ck_factor)
			elseif target == mo.target and dist > length // if far enough away from an adjacent part, pull closer to it
				if dist > 512*mo.scale // (if stuck, become intangible until close enough)
					mo.flags = $ | MF_NOCLIP | MF_NOCLIPHEIGHT
				else
					mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPHEIGHT)
				end
				pullPart(mo, target, length, hdist, player.ck_factor)
			end
			
			target = target.target
		end
		
		mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y) // Face adjacent body part
			
	end
end, MT_CK_PART)

// Body part collision with players

addHook("TouchSpecial", function(part, mo)
	if caterkillercollide.value
	and mo and mo.valid and mo.player // validity checks are such a bore
	and part and part.valid
	and part.tracer and part.tracer.valid and part.tracer.valid and part.tracer.player
	and mo ~= part.tracer // obviously don't fuck with the player whom it's attached to
	and not mo.player.powers[pw_flashing] // While a player is flashing, they should be intangible and immune
	and (not G_BattleGametype() or mo.player.kartstuff[k_bumper] and not mo.player.kartstuff[k_sneakertimer])
		local partsize = part.tracer.player.kartstuff[k_growshrinktimer]
		local mosize = mo.player.kartstuff[k_growshrinktimer]
		partsize = $/abs($ or 1)
		mosize = $/abs($ or 1)
		
		if mosize > partsize // running over a body part while in growth should squish the part's owner
			K_SquishPlayer(part.tracer.player, mo, mo)
		elseif partsize > mosize // likewise, any body part that runs over a player should squish it
			K_SquishPlayer(mo.player, part.tracer, part.tracer)
		elseif mo.player.kartstuff[k_invincibilitytimer] and not part.tracer.player.kartstuff[k_invincibilitytimer] // invinvible players hurt body part owners
			K_SpinPlayer(part.tracer.player, mo, 0, mo)
		elseif part.tracer.player.kartstuff[k_invincibilitytimer] and not mo.player.kartstuff[k_invincibilitytimer] // invincible body parts hurt players
			K_SpinPlayer(mo.player, part.tracer, 0, part.tracer)
		else
			K_KartBouncing(part.tracer, mo) // otherwise just bump each other :D
		end
	end
	return true
end, MT_CK_PART)

// Set up for body parts

addHook("ThinkFrame", do
	for player in players.iterate
		if not player.mo
			continue
		end
		
		local mo = player.mo
		
		if mo.skin ~= SKIN // if not playing as Caterkiller, get rid of the table if necessary and then get out of this thinker asap
			if mo.ck_parts
				mo.ck_parts = nil
			end
			continue
		end
		
		if mo.state >= S_KART_SPIN and mo.state <= S_KART_PAIN // do not spawn body parts if spinning
		or G_BattleGametype() and not player.kartstuff[k_bumper]
			continue
		end
		
		if not mo.ck_parts // set up the table if nonexistent
			mo.ck_parts = {[0] = mo}
		end
		
		if player.name ~= player.ck_name // this is an easter egg, shhhhh
			player.ck_name = player.name
			player.ck_numparts = player.name:lower():find("sandworm") and 5 or 3
			player.ck_factor = player.ck_numparts > 3 and 5 or 7
		end
		
		for i = 1, max(player.ck_numparts, #mo.ck_parts) // check each body part
			if i > player.ck_numparts
				if mo.ck_parts[i] and mo.ck_parts[i].valid and mo.ck_parts[i].health
					killCaterkillerPart(mo.ck_parts[i])
				end
				mo.ck_parts[i] = nil
			elseif not (mo.ck_parts[i] and mo.ck_parts[i].valid and mo.ck_parts[i].health) // if nonexistent/dead, spawn one in!
				local sign = -2*(#player % 2) + 1 // body parts curve clockwise or couterclockwise depending on player number (idk I didn't feel like fucking with rng here)
				local angle = mo.angle - sign*ANGLE_90 - sign*i*ANGLE_45
				
				local x = mo.x + FixedMul(2*mo.radius, cos(mo.angle + sign*ANGLE_90) + cos(angle))
				local y = mo.y + FixedMul(2*mo.radius, sin(mo.angle + sign*ANGLE_90) + sin(angle))
				
				mo.ck_parts[i] = P_SpawnMobj(x, y, mo.z, MT_CK_PART)
				mo.ck_parts[i].target = mo.ck_parts[i - 1]
				mo.ck_parts[i].tracer = mo
				mo.ck_parts[i].flags2 = $ | MF2_DONTDRAW
			end
		end
	end
end)

// btw this character is very sweet and a good friend. they will hug u by gently wrapping their body around you. if you are reading this u deserve many caterkiller hugs.